/*
 * main.c  -- window select plugin for nodoka
 *
 * $Id: main.c,v 1.10 2005/02/10 09:27:24 hos Exp $
 *
 */

#define WINVER 0x0500
#define _WIN32_IE 0x0300
#include <windows.h>
#include <commctrl.h>
#include <psapi.h>
#include <shellapi.h>
#include <tchar.h>
#include "util.h"


static TCHAR *window_class_name = _T("window-select");
static TCHAR *window_title_name = _T("window-select");

static HWND org_selected_hwnd = NULL;
static HWND main_hwnd = NULL;
static HWND list_hwnd = NULL;
static HIMAGELIST list_images = NULL;
static int window_cnt = 0;
static int org_selected_idx = -1;

static HINSTANCE instance = NULL;


#define WM_APP_SELECT (WM_APP + 1)
#define WM_APP_CANCEL (WM_APP + 2)
#define WM_APP_CHANGE_REL (WM_APP + 3)
#define WM_APP_CHANGE (WM_APP + 4)


typedef LRESULT (*msg_proc_t)(HWND, UINT, WPARAM, LPARAM);


static void select_list(int idx)
{
    if(idx < 0) {
        idx += window_cnt;
    }

    ListView_SetItemState(list_hwnd, idx,
                          LVIS_SELECTED | LVIS_FOCUSED,
                          LVIS_SELECTED | LVIS_FOCUSED);
    ListView_EnsureVisible(list_hwnd, idx, FALSE);
}

static void select_list_rel(int idx)
{
    int selected, target;

    if(window_cnt <= 0) {
        return;
    }

    selected = ListView_GetNextItem(list_hwnd, -1, LVNI_SELECTED);
    if(selected < 0) {
        return;
    }

    target = selected + idx;
    while(target < 0) {
        target += window_cnt;
    }
    target %= window_cnt;

    select_list(target);
}

static void set_foreground(void)
{
    int idx;

    idx = ListView_GetNextItem(list_hwnd, -1, LVNI_SELECTED);
    if(idx < 0) {
        return;
    }

    {
        LVITEM item;
        HWND target;

        memset(&item, 0, sizeof(item));
        item.mask = LVIF_PARAM;
        item.iItem = idx;
        item.iSubItem = 0;
        ListView_GetItem(list_hwnd, &item);

        target = (HWND)item.lParam;
        if(! IsWindow(target)) {
            return;
        }

        if(GetWindowLong(target, GWL_STYLE) & WS_MINIMIZE) {
            PostMessage(target, WM_SYSCOMMAND, SC_RESTORE, 0);
        }
        SetForegroundWindow(target);
    }
}


static HICON get_window_icon(HWND hwnd, BOOL use_default)
{
    HICON icon = NULL;

    /* try to get window icon */
    icon = (HICON)SendMessage(hwnd, WM_GETICON, ICON_SMALL, 0);
    if(icon != NULL) {
        return DuplicateIcon(instance, icon);
    }

    /* try to get window class icon */
    icon = (HICON)GetClassLong(hwnd, GCL_HICONSM);
    if(icon != NULL) {
        return DuplicateIcon(instance, icon);
    }

    /* try to get owner window icon */
    {
        HWND owner;

        owner = GetWindow(hwnd, GW_OWNER);
        if(owner != NULL) {
            icon = get_window_icon(owner, FALSE);
        }
    }
    if(icon != NULL) {
        return icon;
    }

    /* try to get window process icon */
    {
        DWORD process_id;
        HANDLE process_handle;

        GetWindowThreadProcessId(hwnd, &process_id);
        process_handle =
            OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ,
                        FALSE, process_id);
        if(process_handle != NULL) {
            HMODULE module[1];
            TCHAR path[MAX_PATH];

            EnumProcessModules(process_handle, module, sizeof(module), NULL);
            if(GetModuleFileNameEx(process_handle, module[0],
                                   path, sizeof(path)) >= 0) {
                icon = ExtractIcon(instance, path, 0);
            }

            CloseHandle(process_handle);
        }
    }
    if(icon != NULL) {
        return icon;
    }

    /* default icon */
    if(use_default) {
        icon = DuplicateIcon(instance, LoadIcon(NULL, IDI_WINLOGO));
    } else {
        icon = NULL;
    }
    return icon;
}

static LPTSTR get_window_name(HWND hwnd)
{
    static TCHAR title[256], class[256];

    memset(title, 0, sizeof(title));
    memset(class, 0, sizeof(class));

    GetWindowText(hwnd, title, 256);
    GetClassName(hwnd, class, 256);

    if(_tcscmp(class, _T("Shell_TrayWnd")) == 0) {
        return _T(" (taskbar) ");
    }
    if(_tcscmp(class, _T("Progman")) == 0) {
        return _T(" (desktop) ");
    }

    if(title[0] != 0) {
        return title;
    }

    if(class[0] != 0) {
        return class;
    }

    return _T(" (no title) ");
}


static BOOL CALLBACK enum_wnd(HWND hwnd, LPARAM lparam)
{
    LONG style, ex_style;
    RECT rc;

    style = GetWindowLong(hwnd, GWL_STYLE);
    ex_style = GetWindowLong(hwnd, GWL_EXSTYLE);
    GetClientRect(hwnd, &rc);

    if((GetParent(hwnd) != NULL && !(style & WS_POPUPWINDOW)) ||
       !(style & WS_VISIBLE) ||
       (style & WS_DISABLED) ||
       ((rc.left == rc.right && rc.top == rc.bottom) &&
         !(style & WS_MINIMIZE))) {
        return TRUE;
    }

    {
        LVITEM item;

        memset(&item, 0, sizeof(item));
        item.mask = LVIF_PARAM | LVIF_TEXT;
        item.iItem = window_cnt;
        item.lParam = (LPARAM)hwnd;
        item.pszText = get_window_name(hwnd);
        ListView_InsertItem(list_hwnd, &item);

        if(hwnd == org_selected_hwnd) {
            org_selected_idx = window_cnt;
        }
    }

    window_cnt++;

    return TRUE;
}

static void listup_windows(void)
{
    EnumWindows(enum_wnd, 0);
    if(window_cnt <= 0) {
        return;
    }

    list_images = ImageList_Create(GetSystemMetrics(SM_CXSMICON),
                                   GetSystemMetrics(SM_CYSMICON),
                                   ILC_COLOR24 | ILC_MASK,
                                   window_cnt, 0);
    if(list_images == NULL) {
        return;
    }

    {
        int i;
        LVITEM item;
        HWND hwnd;
        HICON icon;

        for(i = 0; i < window_cnt; i++) {
            memset(&item, 0, sizeof(item));
            item.mask = LVIF_PARAM;
            item.iItem = i;
            ListView_GetItem(list_hwnd, &item);

            hwnd = (HWND)item.lParam;

            icon = get_window_icon(hwnd, TRUE);
            ImageList_AddIcon(list_images, icon);
            DestroyIcon(icon);

            memset(&item, 0, sizeof(item));
            item.mask = LVIF_IMAGE;
            item.iItem = i;
            item.iImage = i;
            ListView_SetItem(list_hwnd, &item);
        }

        ListView_SetImageList(list_hwnd, list_images, LVSIL_SMALL);
    }
}


static LRESULT killfocus_proc(HWND hwnd, UINT msg,
                              WPARAM wparam, LPARAM lparam)
{
    PostMessage(main_hwnd, WM_CLOSE, 0, 0);

    return 0;
}

static LRESULT dblclk_proc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    set_foreground();
    PostMessage(main_hwnd, WM_CLOSE, 0, 0);

    return 0;
}

static LRESULT app_select_proc(HWND hwnd, UINT msg,
                               WPARAM wparam, LPARAM lparam)
{
    set_foreground();
    PostMessage(main_hwnd, WM_CLOSE, 0, 0);

    return 0;
}

static LRESULT app_cancel_proc(HWND hwnd, UINT msg,
                               WPARAM wparam, LPARAM lparam)
{
    PostMessage(main_hwnd, WM_CLOSE, 0, 0);

    return 0;
}

static LRESULT app_change_rel_proc(HWND hwnd, UINT msg,
                                   WPARAM wparam, LPARAM lparam)
{
    select_list_rel((int)wparam);

    return 0;
}

static LRESULT app_change_proc(HWND hwnd, UINT msg,
                               WPARAM wparam, LPARAM lparam)
{
    select_list(wparam);

    return 0;
}

static LRESULT create_proc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    RECT rect;

    main_hwnd = hwnd;

    GetClientRect(main_hwnd, &rect);

    list_hwnd = CreateWindow(WC_LISTVIEW, _T(""),
                             WS_VISIBLE | WS_CHILD |
                             LVS_REPORT | LVS_SINGLESEL | LVS_SHOWSELALWAYS |
                             LVS_NOCOLUMNHEADER,
                             0, 0, rect.right, rect.bottom,
                             main_hwnd, NULL, NULL, NULL);
    if(list_hwnd == NULL) {
        return -1;
    }

    ListView_SetExtendedListViewStyle(list_hwnd, LVS_EX_FULLROWSELECT);

    {
        LVCOLUMN col;

        memset(&col, 0, sizeof(col));
        col.mask = LVCF_TEXT | LVCF_WIDTH;
        col.cx = rect.right;
        col.pszText = _T("window");
        ListView_InsertColumn(list_hwnd, 0, &col);
    }

    listup_windows();

    ListView_SetColumnWidth(list_hwnd, 0, LVSCW_AUTOSIZE_USEHEADER);

    {
        DWORD pid = 0;
        int size = 256;
        char mod_name[] = "USER32.DLL";
        void *mod, *proc, *lmod, *lproc;
        void *gmh_proc;
        void *remote_addr = NULL;
        HANDLE ph = NULL, thread = NULL;

        /* force put main_hwnd into foreground */

        GetWindowThreadProcessId(GetForegroundWindow(), &pid);
        ((ph = OpenProcess(PROCESS_VM_OPERATION |
                           PROCESS_VM_WRITE | PROCESS_VM_READ |
                           PROCESS_CREATE_THREAD | PROCESS_QUERY_INFORMATION,
                           FALSE, pid)) != NULL &&
         (remote_addr = VirtualAllocEx(ph, NULL, size, MEM_COMMIT,
                                       PAGE_READWRITE)) != NULL &&

         /* get remote handle of USER32.DLL module */
         WriteProcessMemory(ph, remote_addr,
                            mod_name, sizeof(mod_name), NULL) != 0 &&
         (gmh_proc = GetProcAddress(GetModuleHandleA("KERNEL32.DLL"),
                                    "GetModuleHandleA")) != NULL &&
         (thread = CreateRemoteThread(ph, NULL, 0,
                                      (LPTHREAD_START_ROUTINE)gmh_proc,
                                      remote_addr, 0, NULL)) != NULL &&
         WaitForSingleObject(thread, INFINITE) == WAIT_OBJECT_0 &&
         GetExitCodeThread(thread, (LPDWORD)(void *)&mod) != 0 &&
         CloseHandle(thread) != 0 &&
         (thread = NULL, 1) &&

         /* calc remote proc address of SetWindowHandle */
         (lmod = GetModuleHandleA("USER32.DLL")) != NULL &&
         (lproc = GetProcAddress(lmod, "SetForegroundWindow")) != NULL &&
         (proc = ((char*)mod + ((char *)lproc - (char *)lmod)), 1) &&

         /* call remote SetWindowHandle */
         (thread = CreateRemoteThread(ph, NULL, 0,
                                      (LPTHREAD_START_ROUTINE)proc,
                                      main_hwnd, 0, NULL)) != NULL &&
         WaitForSingleObject(thread, INFINITE) == WAIT_OBJECT_0 &&
         CloseHandle(thread) != 0 &&
         (thread = NULL, 1)

         ) || SetForegroundWindow(main_hwnd); /* if failed */

        (thread != NULL && CloseHandle(thread));
        (ph != NULL &&
         (remote_addr != NULL && VirtualFreeEx(ph, remote_addr, size,
                                               MEM_RELEASE)),
         CloseHandle(ph));

        SetFocus(list_hwnd);
    }

    return 0;
}

static LRESULT close_proc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    DestroyWindow(main_hwnd);
    main_hwnd = NULL;

    ImageList_Destroy(list_images);

    return 0;
}

static LRESULT destroy_proc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    return 0;
}

static LRESULT size_proc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    if(list_hwnd != NULL) {
        MoveWindow(list_hwnd, 0, 0, LOWORD(lparam), HIWORD(lparam), TRUE);
        ListView_SetColumnWidth(list_hwnd, 0, LVSCW_AUTOSIZE_USEHEADER);
    }

    return 0;
}

static LRESULT notify_proc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    msg_proc_t proc;
    NMHDR *nmh;

    static struct uint_ptr_pair msg_proc[] = {
        {NM_KILLFOCUS, killfocus_proc},
        {NM_DBLCLK, dblclk_proc},

        {0, NULL}
    };

    nmh = (NMHDR *)lparam;

    proc = (msg_proc_t)assq_pair(msg_proc, nmh->code, NULL);
    if(proc == NULL) {
        return 0;
    }

    return proc(hwnd, msg, wparam, lparam);
}

static LRESULT default_proc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    return DefWindowProc(hwnd, msg, wparam, lparam);
}

static LRESULT CALLBACK wnd_proc(HWND hwnd, UINT msg,
                                 WPARAM wparam, LPARAM lparam)
{
    msg_proc_t proc;

    static struct uint_ptr_pair msg_proc[] = {
        {WM_CREATE, create_proc},
        {WM_DESTROY, destroy_proc},
        {WM_NOTIFY, notify_proc},
        {WM_CLOSE, close_proc},
        {WM_SIZE, size_proc},
        {WM_APP_CANCEL, app_cancel_proc},
        {WM_APP_SELECT, app_select_proc},
        {WM_APP_CHANGE_REL, app_change_rel_proc},
        {WM_APP_CHANGE, app_change_proc},

        {0, NULL}
    };

    proc = (msg_proc_t)assq_pair(msg_proc, msg, default_proc);

    return proc(hwnd, msg, wparam, lparam);
}


static HWND create_window(void)
{
    HWND hwnd;
    int ret;

    {
        WNDCLASSEX wc;

        memset(&wc, 0, sizeof(wc));
        wc.cbSize = sizeof(wc);
        wc.style = 0;
        wc.lpfnWndProc = wnd_proc;
        wc.cbClsExtra = 0;
        wc.cbWndExtra = 0;
        wc.hInstance = instance;
        wc.hIcon = NULL;
        wc.hCursor = NULL;
        wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
        wc.lpszMenuName = NULL;
        wc.lpszClassName = window_class_name;
        wc.hIconSm = NULL;

        ret = RegisterClassEx(&wc);
        if(ret == 0) {
            return NULL;
        }
    }

    {
        RECT rc;
        int x, y, w, h;

        {
            MONITORINFO mi;
            mi.cbSize = sizeof(mi);
            GetMonitorInfo(MonitorFromWindow(org_selected_hwnd,
                                             MONITOR_DEFAULTTONEAREST), &mi);
            rc = mi.rcWork;
        }

        w = 400;
        h = 200;
        x = (rc.left + rc.right - w) / 2;
        y = (rc.top + rc.bottom - h) / 2;

        hwnd = CreateWindowEx(WS_EX_TOPMOST | WS_EX_CLIENTEDGE |
                              WS_EX_TOOLWINDOW,
                              window_class_name, window_title_name,
                              WS_POPUP | WS_DLGFRAME,
                              x, y, w, h,
                              NULL, NULL, NULL, NULL);
    }

    return hwnd;
}


void WINAPI nodokastart(const char *param)
{
    org_selected_hwnd = GetForegroundWindow();

    main_hwnd = create_window();
    if(main_hwnd == NULL) {
        goto func_end;
    }

    ShowWindow(main_hwnd, SW_SHOW);
    if(org_selected_idx >= 0) {
        select_list(org_selected_idx);
    }

    {
        MSG msg;
        int ret;

        while(main_hwnd != NULL) {
            ret = GetMessage(&msg, NULL, 0, 0);
            if(ret <= 0) {
                break;
            }

            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    }

  func_end:
    UnregisterClass(window_class_name, instance);
}


BOOL APIENTRY DllMain(HANDLE module, DWORD reason, LPVOID reserve)
{
    instance = (HINSTANCE)module;

    return TRUE;
}
